import pandas as pd


def parse_attributes(attr: str):
    l = []
    for pair in attr.split(";"):
        if len(pair) > 0:
            k, v = pair.split('=')
        l.append([k.lower(), v])
    return dict(l)


def get_annotation_region(annotation_file: str):
    annot = pd.read_csv(annotation_file,
                        sep='\t',
                        header=None,
                        names=["id", "source", "feature", "start", "end",
                               "score", "strand", "phase", "attributes"],
                        comment='#')
    regions = {}
    for _, row in annot[annot["feature"] == "CDS"].iterrows():
        infos = parse_attributes(row["attributes"])
        if "gene" in infos:
            name = infos["gene"].lower()
        elif "product" in infos:
            name = infos["product"].lower()
        elif "pseudo" in infos and infos["pseudo"] == "true":
            continue
        else:
            print("No label: ", row)
            exit(1)

        known_name_of_1ab = {"1a", "1b", "1ab",
                             "replicase", "polymerase", "polyprotein", "pol"}
        if any(n in name.lower() for n in known_name_of_1ab) or name.lower() == '1' or name.lower() == 'orf1':
            continue

        if name in convert:
            name = convert[name]
        if name.startswith("hypothetical protein"):
            name = name[20:].strip() + '?'
        if name.startswith("non-structural protein"):
            name = "ns" + name[22:].strip()
        if name.startswith("nonstructural protein"):
            name = "ns" + name[21:].strip()
        if name.startswith("putative"):
            name = name[8:].strip() + '??'
        if name.endswith("protein"):
            name = name[:-7].strip()
        if name.startswith("protein"):
            name = name[7:].strip()
        if name.startswith("orf"):
            name = name[3:].strip()
        if name in regions:
            # take the union region if already overlapping a 
            a = regions[name][0]
            b = regions[name][1]
            c = row["start"] - 1
            d = row["end"] - 1
            if (
                (a <= d and c <= b and
                (min(b, d) - max(a, c)) / (max(b, d) - min(a, c)) >= 0.1)
            ):
                regions[name] = (min(a, c), max(b, d))
            else:
                while name in regions:
                    name += '#'
                regions[name] = (row["start"] - 1, row["end"] - 1)
        else:
            regions[name] = (row["start"] - 1, row["end"] - 1)
    return regions

convert = {
    '3A protein': '3a',
    '3a protein': '3a',
    '5a protein#2': '5a#2',
    '7A': '7a',
    '7B': '7b',
    'Membrane protein': 'M',
    'Spike': "S",
    'Spike (S) glycoprotein': "S",
    'e': "E",
    'envelop protein': "E",
    'envelop protein (E)': "E",
    'envelope': "E",
    'envelope E protein': "E",
    'envelope protein': "E",
    'hemagglutinin esterase': "HE",
    'hemagglutinin-esterase': "HE",
    'hemagglutinin-esterase glycoprotein': "HE",
    'hemagglutinin-esterase glycoprotein (HE)': "HE",
    'hemagglutinin-esterase protein': "HE",
    'hypothetical protein#2': "?#2",
    'm': "M",
    'membrance glycoprotein': "M",
    'membrane': "M",
    'membrane M protein': "M",
    'membrane glycoprotein': "M",
    'membrane glycoprotein M': "M",
    'membrane protein': "M",
    'membrane protein (M)': "M",
    'n': "N",
    'nonstructural protein#2': "NS#2",
    'nucleocapsid': "N",
    'nucleocapsid N protein': "N",
    'nucleocapsid phosphoprotein': "N",
    'nucleocapsid protein': "N",
    'nucleocapsid protein (N)': "N",
    'nucleopcapsid N protein': "N",
    'nucleoprotein': "N",
    'nucloecapsid protein': "N",
    'putative internal (I) ORF': "I?",
    's': "S",
    'small envelope E protein': "small E",
    'small envelope protein': "small E",
    'small membrane protein': "small M",
    'spike': "S",
    'spike S glycoprotein': "S",
    'spike glycoprotein': "S",
    'spike glycoprotein (S)': "S",
    'spike protein': "S",
    'spike surface glycoprotein': "S"
}



from dataclasses import dataclass
from typing import List, Sequence, Tuple
import json
from .util import render_color_seq
from .heuristic import guess_orf1ab
from .MWIS import Interval
import numpy as np
from pytablewriter import MarkdownTableWriter
from pytablewriter.style import Style
import sys


@dataclass
class BodyCore:
    ORF: str
    interval_start: int
    interval_len: int
    weight: float
    core_start: int
    core_len:int
    leader_start: int
    leader_end: int
    ORF_start: int
    ORF_len: int
    score: float
    index: int
    align: str


class Result:
    """Single result, contain a TRS alignment and induced genes"""
    bodys: List[BodyCore]

    def __init__(self, leader_core_start, window, intervals: List[Interval], offset, max_weight, compact, ref, regions):
        self.leader_core_start = int(leader_core_start)
        self.leader_core_len = int(window)
        self.leader_core_seq = ref[leader_core_start:leader_core_start+window]
        min_core_s = min(i.ya for i in intervals)
        max_core_e = max(i.yb for i in intervals)
        self.TRS_L_start = min_core_s
        self.TRS_L_len = max_core_e - min_core_s + 1
        self.body_range_start = int(offset)
        self.body_range_len = int(len(ref) - offset)
        self.n_intervals = len(intervals)
        self.weight = float(max_weight)
        self.compact = float(compact)
        if regions is None:
            regions = dict()

        self.bodys = []
        canon = set(regions)
        recall = set()
        for i in intervals:
            for orf, (start, end) in regions.items():
                if (abs(i.orf_start + offset - start) <= 3 and
                    abs(i.orf_end + offset - end) <= 3
                ):
                    assign = orf
                    recall.add(orf)
                    break
            else:
                assign = ""
            align = '-'*(i.ya - min_core_s) + ref[i.start+offset:i.end+offset+1] + '-'*(max_core_e - i.yb)
            self.bodys.append(BodyCore(
                assign,
                int(i.xa + offset),
                int(i.xb - i.xa + 1),
                i.w,
                int(i.start+offset),
                int(i.end - i.start + 1),
                int(i.ya),
                int(i.yb),
                int(i.orf_start + offset),
                int(i.length),
                i.score,
                i.idx,
                align
            ))
        self.total_ORF = list(canon)
        self.recall_ORF = list(recall)
        self.missing_ORF = list(canon - recall)
        self.TRS_L_seq = ref[min_core_s:max_core_e+1]

    def write_result(self, file=sys.stdout):
        """Write result to stdout (default) or a file"""
        print(f"Leader core seq: {self.leader_core_seq}, "
              f"{self.leader_core_start}-{self.leader_core_start + self.leader_core_len}, "
              f"TRS-L: {self.TRS_L_start}-{self.TRS_L_start + self.TRS_L_len}\n"
              f"weight: {self.weight:.3f}, coverage: {1-self.compact:.5f}, Missing: {self.missing_ORF}, "
              f"min score: {min(body.score for body in self.bodys)}",
              file=file)
        header = ["ORF", "interval start-end", "interval len", "weight",
                "core start-end", "leader start-end", "ORF start-end", "score", "ORF len", "index", render_color_seq(self.TRS_L_seq)]
        table = []
        for body in self.bodys:
            # TODO: fix hard-coded index
            table.append([
                body.ORF,
                f"{body.interval_start + 1}-{body.interval_start + body.interval_len + 1}",
                body.interval_len,
                body.weight,
                f"{body.core_start + 1}-{body.core_start + body.core_len + 1}",
                f"{body.leader_start + 1}-{body.leader_end + 1}",
                f"{body.ORF_start + 1}-{body.ORF_start + body.ORF_len + 1}",
                body.score, body.ORF_len, body.index,
                render_color_seq(body.align)
            ])
        writer = MarkdownTableWriter(
            headers=header,
            value_matrix=table,
            margin=1,
        )
        writer.column_styles = [Style(align="center"), Style(align="center"), Style(),  Style(),  Style(align="center"),
                                Style(align="center"),  Style(),  Style(),  Style(),  Style()]
        writer.stream = file
        writer.write_table()

    def to_json(self) -> str:
        """Dump result to a json string"""
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)


class ResultEncoder(json.JSONEncoder):
    """Encoder for class `Results`"""
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, Result) or isinstance(obj, Results) or isinstance(obj, BodyCore):
            return obj.__dict__
        else:
            return super().default(obj)


@dataclass
class Results:
    """
    Collection of results from a genome, with additional meta info
    """
    results: List[Result]
    name: str
    description: str
    genbank: dict
    ORF1ab: Tuple[int]
    sequence: str
    is_corsid_a: bool
    
    def write_result(self, file=sys.stdout):
        for res in self.results:
            res.write_result(file=file)

    def to_json(self):
        return json.dumps(self, cls=ResultEncoder)

    def to_gff(self) -> str:
        """Convert result to GFF3 format"""
        gff = ["##gff-version 3.1.26"]
        gff.append(f"##sequence-region {self.name} {1} {len(self.sequence)}")
        gff.append(f"{self.name}\t.\tgene\t{self.ORF1ab[0] + 1}\t{self.ORF1ab[3] + 3}\t.\t"
                   f"+\t.\tID=gene-1;Name=ORF1;gene_biotype=protein_coding")
        gff.append(f"{self.name}\t.\tCDS\t{self.ORF1ab[0] + 1}\t{self.ORF1ab[1] + 1}\t.\t"
                   f"+\t.\tID=cds-1a;Parent=gene-1;Name=ORF1a")
        gff.append(f"{self.name}\t.\tCDS\t{self.ORF1ab[2] + 1}\t{self.ORF1ab[3] + 3}\t.\t"
                   f"+\t.\tID=cds-1b;Parent=gene-1;Name=ORF1b")
        for i, body in enumerate(self.results[0].bodys):
            if body.ORF:
                gene_id = f"gene-{body.ORF}"
            else:
                gene_id = f"putative-gene-{i+2}"
            attributes = ["ID="+gene_id, "gene_biotype=protein_coding"]
            gff.append(f"{self.name}\t.\tgene\t{body.ORF_start + 1}\t{body.ORF_start + body.ORF_len + 1}\t.\t"
                       f"+\t.\t{';'.join(attributes)}")

            attributes = [f"ID=cds-{i+2}", f"Parent={gene_id}"]
            if body.ORF:
                attributes.append(f"Name={body.ORF}")
                attributes.append(f"gene={body.ORF}")
            else:
                attributes.append(f"Name=protein-{i+2}")
                attributes.append(f"gene=ORF{i+2}")
            gff.append(f"{self.name}\t.\tCDS\t{body.ORF_start + 1}\t{body.ORF_start + body.ORF_len + 1}\t.\t"
                       f"+\t0\t{';'.join(attributes)}")
        return '\n'.join(gff)

@dataclass
class Solution:
    """
    A solution for the MWIS problem, including a TRS-L, and several TRS-Bs.
    """
    weights: List[int]
    intervals: List[list]
    offset: int

    def __post_init__(self):
        if len(self.weights) != len(self.intervals):
            raise RuntimeError(f"Weights and intervals are not of the same length: {len(self.weights)} != {len(self.intervals)}") # TODO: fix the wording
        self.weights = np.array(self.weights)

    def argsort(self, is_lex=False, compact=None) -> np.ndarray:
        """Sort results

        Args:
            is_lex (bool, optional): set to True to sort in lexicographical order of (compact, weights, min alignment score). Defaults to False.
            compact (list, optional): list of compact score. Defaults to None.

        Returns:
            np.ndarray: index in sorted order
        """
        if is_lex and compact is not None:
            min_score = np.array([min(i.score for i in sol) if sol is not None and len(sol) > 0 else 0 for sol in self.intervals])
            return np.lexsort((-min_score, -self.weights, compact))
        else:
            return np.argsort(self.weights)[::-1]

    def serialize_results(self, ref: str, window: int, name: str, description: str,
                          annotation=None, is_lex=False, compact=None, is_corsid_a=False) -> Results:
        """Convert solution to `Results` in order to save

        Args:
            ref (str): reference genome
            window (int): window length
            name (str): name of the genome
            description (str): description of the virus
            annotation (dict, optional): genes in the annotation file. Defaults to None.
            is_lex (bool, optional): set to True to enable lexicographical sorting, need compact as well. Defaults to False.
            compact (ndarray, optional): list of genome coverage. Defaults to None.
            is_corsid_a (bool, optional): is the solution created using CORSID-A. Defaults to False.

        Returns:
            Results: collection of TRS alignments
        """
        results = []
        idx = self.argsort(is_lex, compact)

        for pos in idx:
            if self.intervals[pos] is None or len(self.intervals[pos]) == 0:
                continue
            results.append(Result(pos, window, self.intervals[pos], self.offset, self.weights[pos], compact[pos], ref, annotation))
        start_1a, end_1a, end_1b = guess_orf1ab(ref)
        return Results(results, name, description, annotation, (start_1a, end_1a, end_1a, end_1b), ref, is_corsid_a)

    def get_compact_score(self, ref: str, orf1_end: int, offset: int=0) -> List[float]:
        """Return genome coverage of solutions

        Args:
            ref (str): reference genome
            orf1_end (int): the position of ORF1ab end

        Returns:
            List[float]: array of genome coverage
        """
        compact_score = []
        total_length = len(ref) - orf1_end
        for i, sol in enumerate(self.intervals):
            if sol is None:
                compact_score.append(1)
            else:
                cov = np.zeros(total_length)
                for intv in sol:
                    cov[max(0, intv.orf_start + offset): intv.orf_end + offset + 1] += 1
                compact_score.append((cov==0).mean())
        return np.array(compact_score)



import sys
import pysam
import argparse
import numpy as np
from .util import (
    make_score_func,
    get_description,
    get_name,
)
from .solution import Solution
from .heuristic import (
    guess_orf1ab,
    predict_ORFs,
)
from .MWIS import (
    Interval,
    MWIS,
)
from .annotation import get_annotation_region
from typing import List, Dict, Tuple
from tqdm import tqdm

# Default values
WINDOW = 7
TAU_MAX = 7
TAU_MIN = 2
SHRINK = 0.05


def remove_partial(lefts: List[List[Interval]],
                   rights: List[List[Interval]],
                   partial: List[Interval]):
    """Remove intervals overlap with partial solution from lefts and rights

    Args:
        lefts (List[List[Interval]]): intervals indexed by left endpoints
        rights (List[List[Interval]]): intervals indexed by right endpoints
        partial (List[Interval]): partial solution
    """
    for i in range(len(lefts)):
        for intv in partial:
            if intv.xa <= i <= intv.xb:
                lefts[i] = []
            else:
                lefts[i] = [l for l in lefts[i] if (l.len_x_overlap(intv.xa, intv.xb) == 0 and
                                                    l.len_x_overlap_orf(0, 200) < l.orf_len and
                                                    l.len_x_overlap_orf(intv.orf_start, intv.orf_end) < l.orf_len)]
    for i in range(len(rights)):
        for intv in partial:
            if intv.xa <= i <= intv.xb:
                rights[i] = []
            else:
                rights[i] = [r for r in rights[i] if (r.len_x_overlap(intv.xa, intv.xb) == 0 and
                                                      r.len_x_overlap_orf(0, 200) < r.orf_len and
                                                      r.len_x_overlap_orf(intv.orf_start, intv.orf_end) < r.orf_len)]


def semi_smith_waterman(s1: str,
                        s2: str,
                        window: int,
                        match=1,
                        mismatch=-1,
                        tau_min=TAU_MIN,
                        tau_max=TAU_MAX,
                        orf_thr=100,
                        shrink=0.05):
    """Modified Smith-Waterman

    The DP table for each window start is stitched together by the local alignment table and the 
    diagonal table at the row `window_start`.

    Args:
        s1 (str): leader sequence
        s2 (str): body sequence
        window (int): window length
        match (int, optional): matching score. Defaults to 1.
        mismatch (int, optional): mismatch score. Defaults to -1.
        tau_min (int, optional): minimum tau (min score). Defaults to TAU_MIN.
        tau_max (int, optional): maximum tau (min score). Defaults to TAU_MAX.
        orf_thr (int, optional): ORF length threshold. Defaults to 100.
        shrink (float, optional): ORF shrinking percentage. Defaults to 0.05.

    Returns:
        Tuple: maximum scores, and intervals
    """
    s1 = s1.upper()
    s2 = s2.upper()
    len1 = len(s1)
    len2 = len(s2)
    score_f = make_score_func(match, mismatch)
    next_start, possible_trs = predict_ORFs(s2)

    # Prepare diagonal scores for i >= window start
    diag = np.zeros((len1 + 1, len2 + 1), dtype=int)
    for j in range(len2):
        for i in range(len1):
            diag[i, j] = diag[i - 1, j - 1] + score_f(s1[i], s2[j])

    # Prepare local alignment scores for i < window start
    local = np.zeros((len1 + 1, len2 + 1), dtype=int)
    origin = np.zeros((len1 + 1, len2 + 1, 2), dtype=int)
    origin[0, :, 0] = 0
    origin[0, :, 1] = np.arange(0, len2 + 1)
    origin[-1, :, 0] = 0
    origin[-1, :, 1] = np.arange(1, len2 + 2)
    origin[:, 0, 0] = np.arange(0, len1 + 1)
    origin[:, 0, 1] = 0
    origin[:, -1, 0] = np.arange(1, len1 + 2)
    origin[:, -1, 1] = 0
    for j in range(len2):
        for i in range(len1):
            new_score = local[i - 1, j - 1] + score_f(s1[i], s2[j])
            if new_score > 0:
                local[i, j] = new_score
                origin[i, j, :] = origin[i - 1, j - 1, :]
            else:
                local[i, j] = 0
                origin[i, j, :] = [i, j]

    # MWIS
    score_sweep = {i: [0] * (len1 - window) for i in range(tau_min, tau_max+1)}
    intervals_sweep = {i: [None] * (len1 - window) for i in range(tau_min, tau_max+1)}
    for window_start in tqdm(range(0, len1 - window)):
        d_intervals = {}
        # Iterate along diagonal lines
        for idx_diag in range(1 - window_start, len2 - window_start - window + 1):
            prev = window_start - 1
            # delta = the shift of the diagonal line caused by stitching with
            # the local alignment table
            if window_start == 0:
                delta = 0
            else:
                delta = local[prev, idx_diag + prev] - diag[prev, idx_diag + prev]
            next_orf_start = next_start[idx_diag + window_start + window - 3]
            if next_orf_start is None:
                continue
            next_orf_end = possible_trs[next_orf_start]
            len_orf = next_orf_end - next_orf_start
            
            # try once more if length is not enough
            if len_orf < orf_thr:
                try:
                    next_orf_start = next_start[next_orf_start+3]
                except IndexError:
                    continue
                if next_orf_start is None:
                    continue
                next_orf_end = possible_trs[next_orf_start]
                len_orf = next_orf_end - next_orf_start
                if len_orf < orf_thr:
                    continue

            for i in range(window_start + window - 1, len1):
                j = i + idx_diag
                # Skip lower left triangular region,
                # since it won't contain full window
                if j >= len2:
                    break
                cur_score = diag[i, j] + delta
                if (
                    (cur_score >= 2) and
                    next_start[j] is not None
                ):
                    weight = float(len_orf) + cur_score / 1000
                    if len_orf > orf_thr:
                        y = origin[prev, idx_diag + prev, 0]
                        x = origin[prev, idx_diag + prev, 1]
                        shrink_start = int(next_orf_start + shrink * len_orf)
                        shrink_end = int(next_orf_start + (1-shrink) * len_orf)
                        if (shrink_start, shrink_end) not in d_intervals:
                            d_intervals[shrink_start, shrink_end] = Interval(0, shrink_start, shrink_end, weight,
                                        y + 1, i, x + 1, j, cur_score, len_orf, next_orf_start, next_orf_end)
                        else:
                            old_intv = d_intervals[shrink_start, shrink_end]
                            if cur_score > old_intv.score:
                                d_intervals[shrink_start, shrink_end] = Interval(0, shrink_start, shrink_end, weight,
                                        y + 1, i, x + 1, j, cur_score, len_orf, next_orf_start, next_orf_end)

        lefts = [[] for _ in range(len2)]
        rights = [[] for _ in range(len2)]
        # print(f"m={len(d_intervals)}", end=' ')
        for (l, r), intv in d_intervals.items():
            lefts[l].append(intv)
            rights[r].append(intv)
        idx_intv = 0
        for right in rights:
            for r in right:
                r.idx = idx_intv
                idx_intv += 1
        score, intervals = MWIS(lefts, rights, thr=tau_max)
        score_sweep[tau_max][window_start] = score
        intervals_sweep[tau_max][window_start] = intervals

        if len(intervals) == 0:
            continue
        remove_partial(lefts, rights, intervals)
        for thr in range(tau_max - 1, tau_min-1, -1):
            score, intervals = MWIS(lefts, rights, thr=thr)
            if len(intervals) > 0:
                intervals_sweep[thr][window_start] = merge_solutions(intervals_sweep[thr+1][window_start], intervals)
                score_sweep[thr][window_start] = sum(i.w for i in intervals_sweep[thr][window_start])
                remove_partial(lefts, rights, intervals_sweep[thr][window_start])
            else:
                intervals_sweep[thr][window_start] = intervals_sweep[thr+1][window_start]
                score_sweep[thr][window_start] = score_sweep[thr+1][window_start]
    return score_sweep, intervals_sweep


def merge_solutions(sol1: List[Interval], sol2: List[Interval]):
    """Merge 2 solutions and sort them

    Args:
        sol1 (List[Interval]): Solution 1
        sol2 (List[Interval]): Solution 2

    Returns:
        List[Interval]: Merged sorted solution
    """
    assert len(sol1) > 0
    assert len(sol2) > 0
    i = 0
    j = 0
    sol = []
    while True:
        if sol2[j].xa <= sol1[i].xa:
            sol.append(sol1[i])
            i += 1
        else:
            sol.append(sol2[j])
            j += 1
        if i >= len(sol1):
            sol.extend(sol2[j:])
            break
        if j >= len(sol2):
            sol.extend(sol1[i:])
            break
    return sol


def corsid(ref: str,
           annotation: Dict[str, Tuple],
           description: str,
           name: str,
           window: int,
           tau_min: int=TAU_MIN,
           tau_max: int=TAU_MAX,
           mismatch: int=-1,
           shrink: float=0.05):
    """Entrypoint of CORSID

    Args:
        ref (str): reference genome
        annotation (Dict[str, Tuple]): genes in the annotation file
        description ([type]): description string in the annotation file
        name ([type]): name of the genome
        window ([type]): sweeping window length
        tau_min (int, optional): minimum tau (min score). Defaults to TAU_MIN.
        tau_max (int, optional): maximum tau (min score). Defaults to TAU_MAX.
        mismatch (int, optional): mismatch score. Defaults to -1.
        shrink (float, optional): fraction of positions that may overlap between consecutive genes. Defaults to 0.05.

    Returns:
        Tuple[Results, List[float]]: Results and compact scores for each result.
    """
    leader_end, _, orf1ab_end = guess_orf1ab(ref)
    leader_end = min(leader_end, 500)
    offset = orf1ab_end - 200
    leader = ref[:leader_end].replace('N', '-')
    max_weight_sweep, intervals_sweep = semi_smith_waterman(leader, ref[offset:],
                                                            window, mismatch=mismatch, 
                                                            tau_min=tau_min, tau_max=tau_max,
                                                            orf_thr=100, shrink=shrink)
    solution1 = {s: Solution(max_weight_sweep[s], intervals_sweep[s], offset) for s in max_weight_sweep}
    compact_score = solution1[tau_min].get_compact_score(ref, solution1[tau_min].offset)
    result1 = solution1[tau_min].serialize_results(ref,
                                          window,
                                          name,
                                          description,
                                          annotation,
                                          is_lex=True,
                                          compact=compact_score)

    return result1, compact_score


def main():
    parser = argparse.ArgumentParser()

    required = parser.add_argument_group('required arguments')
    required.add_argument("-f", "--fasta", required=True,
                          type=str, help="FASTA genome file")

    parser.add_argument("-g", "--gff", type=str, help="GFF annotation file")
    parser.add_argument("-n", "--name", type=str,
                        help="sample name",
                        default=None)
    parser.add_argument("-o", "--output", type=str, help="output json file name")
    parser.add_argument("-r", "--output-orf", type=str,
                        help="output identified ORFs (FASTA), only contains the first solution")
    parser.add_argument("-3", "--output-gff3", type=str,
                        help="output identified ORFs (FASTA), only contains the first solution")
    parser.add_argument("-w", "--window", type=int,
                        help=f"length of sliding window [{WINDOW}]",
                        default=WINDOW)
    parser.add_argument("-x", "--mismatch", type=int,
                        help=f"mismatch score [-2]",
                        default=-2)
    parser.add_argument("-t", "--tau_min", type=int,
                        help=f"minimum matching score threshold [{TAU_MIN}]",
                        default=TAU_MIN)
    parser.add_argument("-T", "--tau_max", type=int,
                        help=f"maximum matching score threshold [{TAU_MAX}]",
                        default=TAU_MAX)
    parser.add_argument("--shrink", type=float,
                        help=f"fraction of positions that may overlap between consecutive genes [{SHRINK}]",
                        default=SHRINK)
    args = parser.parse_args(None if sys.argv[1:] else ['-h'])

    print(' '.join(sys.argv))

    fasta = pysam.Fastafile(args.fasta)
    ref = fasta.fetch(fasta.references[0])
    if args.gff:
        annotation = get_annotation_region(args.gff)
    else:
        annotation = None
    description = get_description(args.fasta)

    if args.name is None:
        name = get_name(args.fasta)
    else:
        name = args.name

    result, compact_score = corsid(ref,
                                       annotation,
                                       description,
                                       name,
                                       args.window,
                                       args.tau_min,
                                       args.tau_max,
                                       args.mismatch,
                                       args.shrink)

    # Fall back to smaller window if not compact enough
    opt_pos = result.results[0].leader_core_start
    if compact_score[opt_pos] >= 0.2:
        result, _ = corsid(ref,
                               annotation,
                               description,
                               name,
                               args.window - 1,
                               args.tau_min,
                               args.tau_max,
                               args.mismatch,
                               args.shrink)

    result.write_result()

    if args.output:
        with open(args.output, "w") as ofile:
            ofile.write(result.to_json())
        # write GFF3 file
        gff_name = '.'.join(args.output.split('.')[:-1] + ["gff"])
        with open(gff_name, "w") as ofile:
            ofile.write(result.to_gff())

    if args.output_orf:
        with open(args.output_orf, "w") as ofile:
            content = []
            ORFs = result.results[0].bodys
            for i, body in enumerate(ORFs):
                if body.ORF:
                    content.append(f">{body.ORF}")
                else:
                    content.append(f">putative_ORF_{i+1}")
                content.append(ref[body.ORF_start:body.ORF_start + body.ORF_len + 3])
            ofile.write('\n'.join(content))


if __name__ == "__main__":
    main()



import pandas as pd

def get_df_orf(ref):
    """Get as dataframe of all possible ORFs

    Args:
        ref (str): reference genome

    Returns:
        pd.DataFrame: dataframe of all possible ORFs
    """
    possible_orf = []
    start_pos = []
    start_pos.append(ref.find('ATG'))
    while start_pos[-1] != -1:
        start_pos.append(ref.find('ATG', start_pos[-1]+1))

    stop_codons = ["TAA", "TAG", "TGA"]
    stop_pos = []
    for i in range(len(ref)-3):
        if ref[i:i+3] in stop_codons:
            stop_pos.append(i)

    possible_orf = []
    for start in start_pos:
        for stop in stop_pos:
            if stop > start and (stop - start) % 3 == 0:
                possible_orf.append((start, stop))
                break

    df_orf = pd.DataFrame(possible_orf, columns=["start", "stop"])
    df_orf["length"] = df_orf["stop"] - df_orf["start"]
    df_orf.sort_values("start", inplace=True)
    return df_orf.iloc[1:, :]

def guess_orf1ab(ref):
    """Heuristic for finding ORF1ab

    Args:
        ref (str): reference genome

    Returns:
        tuple: start of 1a, end of 1a, end of 1b
    """
    df_orf = get_df_orf(ref)

    length = len(ref)
    founds = []
    stop_codons = ["TAA", "TAG", "TGA"]
    for idx, row in df_orf.iterrows():
        found = []
        for i in range(row["stop"]-1, length-3, 3):
            if ref[i:i+3] in stop_codons:
                found = (row["start"], row["stop"], i)
                break
        else:
            break
        for i in range(row["stop"]-2, length-3, 3):
            if ref[i:i+3] in stop_codons:
                if i - row["start"] > found[2] - found[0]:
                    found = (row["start"], row["stop"], i)
                break
        founds.append(found)
    max_idx = 0
    for i, f in enumerate(founds):
        if f[2] - f[0] > founds[max_idx][2] - founds[max_idx][0]:
            max_idx = i
    return founds[max_idx][0], founds[max_idx][1], founds[max_idx][2]


def predict_ORFs(seq: str):
    """Predict putative ORFs

    Args:
        seq (str): genome

    Returns:
        tuple: the list of next start codons, and corresponding stop codon for each start codon
    """
    length = len(seq)
    start_pos = []
    start_pos.append(seq.find('ATG'))
    while True:
        next_start = seq.find('ATG', start_pos[-1]+1)
        if next_start != -1:
            start_pos.append(next_start)
        else:
            break
    next_start = [None] * length
    pointer = 0
    for i in range(length):
        if i <= start_pos[pointer]+2:
            next_start[i] = start_pos[pointer]
        else:
            pointer += 1
            if pointer < len(start_pos):
                next_start[i] = start_pos[pointer]
            else:
                break

    stop_codons = ["TAA", "TAG", "TGA"]
    stop_pos = []
    for i in range(length-3):
        if seq[i:i+3] in stop_codons:
            stop_pos.append(i)

    next_stop = {}
    for start in start_pos:
        for stop in stop_pos:
            if stop > start and (stop - start) % 3 == 0:
                next_stop[start] = stop
                break
        else:
            next_stop[start] = length - 1

    return next_start, next_stop



from functools import wraps
from time import time

def color_code(base):
    """Unix console color code for nucleotides

    Args:
        base (str): nucleotides (ACGT)

    Returns:
        str: colored nucleotides
    """
    if base == "G":
        return "\x1b[48;5;220mG\x1b[0m"
    elif base == "C":
        return "\x1b[48;5;204mC\x1b[0m"
    elif base == "A":
        return "\x1b[48;5;118mA\x1b[0m"
    elif base == "T":
        return "\x1b[48;5;81mT\x1b[0m"
    else:
        return base


def render_color_seq(seq: str):
    """Colorful base pairs

    Args:
        seq (str): base pairs

    Returns:
        str: base pairs with colors
    """
    return ''.join(map(color_code, seq))


def make_score_func(match=1, mismatch=-1, indel=-1):
    """make a scoring function

    Args:
        match (int, optional): matching score. Defaults to 1.
        mismatch (int, optional): mismatch score. Defaults to -1.
        indel (int, optional): indel score. Defaults to -1.

    Returns:
        function: scoring function
    """
    characters = ['A', 'C', 'G', 'T', 'N', 'R', 'W', 'S',
                  'Y', 'M', 'K', 'V', 'H', 'D', 'B']
    represents = {
        'A': {'A'},
        'C': {'C'},
        'G': {'G'},
        'T': {'T'},
        'N': {'A', 'C', 'G', 'T'},
        'R': {'A', 'G'},
        'W': {'A', 'T'},
        'S': {'C', 'G'},
        'Y': {'C', 'T'},
        'M': {'A', 'C'},
        'K': {'G', 'T'},
        'V': {'A', 'C', 'G'},
        'H': {'A', 'C', 'T'},
        'D': {'A', 'G', 'T'},
        'B': {'C', 'G', 'T'},
    }
    score = {}
    for c in characters:
        score[c] = {k: match if k in represents[c]
                    else mismatch for k in characters}
        score[c]['-'] = indel
    score['-'] = {k: indel for k in characters}
    score['-']['-'] = 0
    return lambda a, b: score[a][b]

def get_description(fasta):
    """get description from FASTA file

    Args:
        fasta (str): path to FASTA file

    Returns:
        str: description
    """
    with open(fasta) as ifile:
        for line in ifile:
            if line.startswith(">"):
                name_split = line.find(' ')
                return line[name_split+1:].strip()

def get_name(fasta):
    """get genome name from FASTA file

    Args:
        fasta (str): path to FASTA file

    Returns:
        str: genome name
    """
    with open(fasta) as ifile:
        for line in ifile:
            if line.startswith(">"):
                name_split = line.find(' ')
                return line[1:name_split].strip()


def timing(f):
    @wraps(f)
    def wrap(*args, **kw):
        ts = time()
        result = f(*args, **kw)
        te = time()
        print(f'func:{f.__name__} took: {te-ts:2.4f} sec')
        return result
    return wrap


import sys
import pysam
import argparse
import numpy as np
from .util import (
    get_name,
    make_score_func,
    get_description,
    timing,
)
from .solution import Solution
from .heuristic import (
    guess_orf1ab,
    predict_ORFs,
)
from .MWIS import (
    Interval,
)
from .annotation import get_annotation_region
from tqdm import tqdm

def get_candidate_region(fasta_file: str, annotation_file: str):
    regions = get_annotation_region(annotation_file)
    fasta = pysam.FastaFile(fasta_file)
    contig = fasta.references[0]
    ref = fasta.fetch(contig)
    _, possible_trs = predict_ORFs(ref)
    candidates = {}
    old_start = 0
    for name, (start, end) in regions.items():
        previous_ATG = ref[:start].rfind("ATG")
        previous_2ATG = ref[:previous_ATG].rfind("ATG")
        if possible_trs[previous_ATG] - previous_ATG < 100:
            candidates[name] = (max(old_start, previous_2ATG), start)
        else:
            candidates[name] = (max(old_start, previous_ATG), start)
        old_start = start
    return candidates

def get_orf_length(seq, starts):
    stop_codons = ["TAA", "TAG", "TGA"]
    stop_pos = []
    for i in range(len(seq)-3):
        if seq[i:i+3] in stop_codons:
            stop_pos.append(i)
    possible_orf = {}
    for start in starts:
        for stop in stop_pos:
            if stop > start and (stop - start) % 3 == 0:
                possible_orf[start] = stop
                break
        else:
            possible_orf[start] = len(seq)
    return possible_orf

def semi_smith_waterman(s1: str, s2: str, genes, window, match=1, mismatch=-1, indel=-1):
    s1 = s1.upper()
    s2 = s2.upper()
    len1 = len(s1)
    len2 = len(s2)
    score = make_score_func(match, mismatch, indel)
    possible_orf = get_orf_length(s2, [start for _, start in genes])

    # Prepare diagonal scores for i >= window start
    diag = np.zeros((len1+1, len2+1), dtype=int)
    for start, end in genes:
        for j in range(start, end):
            for i in range(0, len1):
                diag[i, j] = diag[i-1, j-1] + score(s1[i], s2[j])

    # Prepare local alignment scores for i < window start
    local = np.zeros((len1+1, len2+1), dtype=int)
    origin = np.zeros((len1+1, len2+1, 2), dtype=int)
    origin[0, :, 0] = 0
    origin[0, :, 1] = np.arange(0, len2 + 1)
    origin[-1, :, 0] = 0
    origin[-1, :, 1] = np.arange(1, len2 + 2)
    origin[:, 0, 0] = np.arange(0, len1 + 1)
    origin[:, 0, 1] = 0
    origin[:, -1, 0] = np.arange(1, len1 + 2)
    origin[:, -1, 1] = 0
    for start, end in genes:
        origin[:, start, 0] = np.arange(0, len1 + 1)
        origin[:, start, 1] = start
        origin[:, start-1, 0] = np.arange(0, len1 + 1)
        origin[:, start-1, 1] = start-1
        for j in range(start, end):
            for i in range(0, len1):
                new_score = local[i-1, j-1] + score(s1[i], s2[j])
                if new_score > 0:
                    local[i, j] = new_score
                    origin[i, j, :] = origin[i-1, j-1, :]
                else:
                    local[i, j] = 0
                    origin[i, j, :] = [i, j]

    # dynamic programming
    score_sweep = []
    intervals_sweep = []
    for window_start in tqdm(range(0, len1-window)):
        max_scores = []
        max_intervals = []
        for idx, (start, end) in enumerate(genes):
            max_score = float("-inf")
            max_intv = None
            for i in range(window_start + window - 1, len1):
                # Skip lower left triangular region,
                # since it won't contain full window
                for j in range(start+window+i-window_start-window+1, end):
                    idx_diag = j - i
                    prev = window_start - 1
                    # delta = the shift of the diagonal line caused by stitching with
                    # the local alignment table
                    if window_start == 0:
                        delta = 0
                    else:
                        delta = local[prev, idx_diag + prev] - diag[prev, idx_diag + prev]
                    cur_score = diag[i, j] + delta
                    if cur_score >= 2 and cur_score >= max_score:
                        max_score = cur_score
                        len_orf = possible_orf[end] - end
                        x, y = origin[prev, idx_diag + prev, :]
                        max_intv = Interval(0, y, end, cur_score,
                                            x+1, i, y+1, j, cur_score, len_orf, end, possible_orf[end])
            if max_intv is not None:
                max_scores.append(max_score)
                max_intervals.append(max_intv)
        
        score_sweep.append(sum(max_scores))
        intervals_sweep.append(max_intervals[::-1])
    return score_sweep, intervals_sweep


WINDOW = 7
SCORE = 2

@timing
def corid_annot(ref, regions, annotation, name, description, window, mismatch):
    leader_end, _, orf1ab_end = guess_orf1ab(ref)
    leader_end = min(leader_end, 500)
    genes = list((s-leader_end, e-leader_end) for s, e in regions.values())
    genes = [gene for gene in genes if gene[0] > 0 and gene[1] > 0]
    leader = ref[:leader_end].replace('N', '-')

    score_sweep, intervals_sweep = semi_smith_waterman(leader,
                                                       ref[leader_end:],
                                                       genes,
                                                       window=window,
                                                       mismatch=mismatch)

    solution = Solution(score_sweep, intervals_sweep, leader_end)
    compact_score = solution.get_compact_score(ref, orf1ab_end, offset=leader_end-orf1ab_end)
    result = solution.serialize_results(ref, window, name, description, annotation, compact=compact_score, is_corsid_a=True)

    return result


def main():
    parser = argparse.ArgumentParser()

    required = parser.add_argument_group('required arguments')
    required.add_argument("-f", "--fasta", required=True, type=str, help="FASTA genome file")
    required.add_argument("-g", "--gff", required=True, type=str, help="GFF annotation file")
    
    parser.add_argument("-n", "--name", type=str,
                        help="sample name",
                        default=None)
    parser.add_argument("-o", "--output", type=str, help="output file name")
    parser.add_argument("-w", "--window", type=int,
                        help=f"length of sliding window [{WINDOW}]",
                        default=WINDOW)
    parser.add_argument("-x", "--mismatch", type=int,
                        help=f"mismatch score [-2]",
                        default=-2)
    parser.add_argument("-s", "--score", type=int,
                        help=f"minimum alignment score threshold [{SCORE}]",
                        default=SCORE)
    args = parser.parse_args(None if sys.argv[1:] else ['-h'])

    print(' '.join(sys.argv))

    fasta = pysam.Fastafile(args.fasta)
    ref = fasta.fetch(fasta.references[0])
    regions = get_candidate_region(args.fasta, args.gff)
    annotation = get_annotation_region(args.gff)
    description = get_description(args.fasta)

    if args.name is None:
        name = get_name(args.fasta)
    else:
        name = args.name

    results = corid_annot(ref,
                            regions,
                            annotation,
                            name,
                            description,
                            args.window,
                            args.mismatch)
    results.write_result()

    if args.output:
        with open(args.output, "w") as ofile:
            ofile.write(results.to_json())

if __name__ == "__main__":
    main()



from typing import List, Tuple
from dataclasses import dataclass


@dataclass
class Interval:
    """Closed interval [a, b] on both x and y."""
    idx: int
    xa: int
    xb: int
    w: float
    ya: int
    yb: int
    start: int
    end: int
    score: int
    length: int
    orf_start: int
    orf_end: int

    def __post_init__(self):
        self.orf_len = self.orf_end - self.orf_start + 1

    def __str__(self):
        return f"{self.idx}\t{self.xa}-{self.xb}\t{self.w}\t{self.ya}-{self.yb}\t{self.start}-{self.end}\t{self.orf_start}-{self.orf_end}\t{self.score}\t{self.length}"

    def __repr__(self):
        return f"<{self.idx}|xa={self.xa}, xb={self.xb}, w={self.w}, ya={self.ya}, yb={self.yb}, start={self.start}, end={self.end}, start={self.orf_start}, end={self.orf_end}, score={self.score}, len={self.length}>"

    def __contains__(self, item):
        return self.orf_start <= item.orf_start and item.orf_end <= self.orf_end

    def len_x_overlap(self, start, end):
        """[x_a, x_b] overlap with [start, end]"""
        if start >= end:
            return 0
        elif start > self.xb or self.xa > end:
            return 0
        else:
            return min(self.xb, end) - max(self.xa, start) + 1

    def len_x_overlap_orf(self, start, end):
        """[orf_start, orf_end] overlap with [start, end]"""
        if start >= end:
            return 0
        elif start > self.orf_end or self.orf_start > end:
            return 0
        else:
            return min(self.orf_end, end) - max(self.orf_start, start) + 1


def MWIS(lefts: List[List[Interval]], rights: List[List[Interval]], thr: float=0) -> Tuple[float, List[Interval]]:
    """Maximum weight independent set.

    Intervals should be sorted by their right endpoints.

    Args:
        lefts (List[List[Interval]]): intervals indexed by left endpoints
        rights (List[List[Interval]]): intervals indexed by left endpoints
        thr (int, optional): minimum score threshold. Defaults to 0.

    Returns:
        Tuple[float, List[Interval]]: max weight and the optimal set.
    """
    n_left = sum(len(l) for l in lefts)
    n_right = sum(len(r) for r in rights)
    assert n_left == n_right
    if n_left == 0:
        return 0, []
    n_intv = max([max(l.idx for l in left) for left in lefts if len(left) > 0]) + 1
    sofar = [-1] * n_intv
    map_intervals = {}
    max_sofar = 0
    max_interval = None
    for j in range(len(lefts)):
        left = lefts[j]
        right = rights[j]
        for l in left:
            if l.score >= thr:
                sofar[l.idx] = max_sofar + l.w
        for r in right:
            if r.score >= thr:
                map_intervals[r.idx] = r
                if sofar[r.idx] > max_sofar:
                    max_sofar = sofar[r.idx]
                    max_interval = r

    if max_interval is None:
        return 0, []

    intervals = [max_interval]
    max_w = max_sofar - max_interval.w
    while True:
        for j in range(intervals[-1].idx-1, -1, -1):
            if abs(sofar[j] - max_w) <= 1e-9 and j in map_intervals and map_intervals[j].xb < intervals[-1].xa:
                intervals.append(map_intervals[j])
                max_w -= map_intervals[j].w
                break
        else:
            break
        if j == 0:
            break
    return max_sofar, intervals



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setuptools.setup(
    name='corsid',
    packages=["corsid"],
    description="Core Sequence Identifier",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version='0.1.2',
    url='http://github.com/elkebir-group/CORSID',
    author='Chuanyi Zhang',
    author_email='chuanyi5@illinois.edu',
    license='MIT',
    python_requires='>=3.7',
    scripts=[
        'scripts/corsid',
        'scripts/corsid_a',
    ],
    install_requires=[
        "numpy",
        "pysam",
        "pandas",
        "pytablewriter",
        "tqdm",
    ],
)